
/*
    JoyGen.scx, v0.1
    Written by Christophe McKeon, Month, day, year
    Includong a great deal ripped right off the MacUgens, props to J. McCartney
    License: SuperCollider Server License, see below
*/

/*
	SuperCollider real time audio synthesis system
    Copyright (c) 2002 James McCartney. All rights reserved.
	http://www.audiosynth.com

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <pthread.h>

#define JOYGEN_SDL_JOYSTICK_API 1

#if SC_DARWIN
    #include <unistd.h>
    #define JOYGEN_JOYSTICK_API JOYGEN_SDL_JOYSTICK_API
#elif SC_LINUX
    #include <time.h>
    #define JOYGEN_JOYSTICK_API JOYGEN_SDL_JOYSTICK_API
#endif

#include "SC_PlugIn.h"
#include "SDL.h"

// these correspond exactly with the sdl hat #defines
// they are here for joystick api independence though
#define JOYGEN_HAT_CENTERED		0x00
#define JOYGEN_HAT_UP			0x01
#define JOYGEN_HAT_RIGHT		0x02
#define JOYGEN_HAT_DOWN			0x04
#define JOYGEN_HAT_LEFT			0x08
#define JOYGEN_HAT_RIGHTUP		(JOYGEN_HAT_RIGHT|JOYGEN_HAT_UP)
#define JOYGEN_HAT_RIGHTDOWN		(JOYGEN_HAT_RIGHT|JOYGEN_HAT_DOWN)
#define JOYGEN_HAT_LEFTUP		(JOYGEN_HAT_LEFT|JOYGEN_HAT_UP)
#define JOYGEN_HAT_LEFTDOWN		(JOYGEN_HAT_LEFT|JOYGEN_HAT_DOWN)

namespace {

    const int UPDATE_TIME_USECS = 20000; //17000
    
    #if SC_LINUX
    // e.g. if linux, is there a better way to #if for it?
        struct timespec gRequestedTime , gRemainingTime;
    
        gRequestedTime.tv_sec = 0;
        gRequestedTime.tv_nsec = UPDATE_TIME_USECS * 1000;
        
        const int MAX_BALLS = 2;
    #endif
    
    const int MAX_BUTTONS = 24;
    const int MAX_AXES = 8;
    const int MAX_HATS = 4;
    
}

////////////////////////////////////////////////////////////////////////////////////////

#if SC_LINUX
struct BallVals {
    float x, y;
};
#endif

struct JoyStick {
    
    int numButtons;
    int numAxes;
    int numHats;

    int buttonVals[MAX_BUTTONS];
    float axisVals[MAX_AXES];
    int hatVals[MAX_HATS];
    #if SC_LINUX
    int numBalls;
    BallVals ballValsX[MAX_BALLS];
    #endif
};

////////////////////////////////////////////////////////////////////////////////////////

namespace {
    InterfaceTable *ft;
    
    #if JOYGEN_JOYSTICK_API == JOYGEN_SDL_JOYSTICK_API
    SDL_Joystick ** gSDLSticks;
    #endif
    
    JoyStick * gLocalSticks;
    int gNumJoySticks;
}

////////////////////////////////////////////////////////////////////////////////////////

struct JoyGen : public Unit {
    int m_id, m_index;
    float m_y1, m_b1, m_lag;
};

struct JoyButton : public JoyGen {
    bool m_state;
    int m_prevVal;
};

struct JoyHat : public JoyGen {
    int m_prevPos;
};

inline void initJoyGen( JoyGen * unit ) {
    unit->m_id = (int)ZIN0(0);
    unit->m_index = (int)ZIN0(1);
    unit->m_b1 = 0.f;
    unit->m_lag = 0.f;
}

////////////////////////////////////////////////////////////////////////////////////////	

extern "C" {
    void JoyButton_Ctor( JoyButton * unit );
    void JoyAxis_Ctor( JoyGen * unit );
    void JoyHat_Ctor( JoyHat * unit );
    

    void JoyButton_next( JoyButton * unit, int inNumSamples );
    void JoyButton_next_t( JoyButton * unit, int inNumSamples );
    void JoyAxis_next( JoyGen * unit, int inNumSamples );
    void JoyHat_next( JoyHat * unit, int inNumSamples );
    void JoyHat_next_t( JoyHat * unit, int inNumSamples );
}

////////////////////////////////////////////////////////////////////////////////////////	

void JoyButton_next( JoyButton * unit, int inNumSamples ) {
    float minval = ZIN0(2);
    float maxval = ZIN0(3);
    float lag = ZIN0(4);
    
    float y0;
    float y1 = unit->m_y1;
    float b1 = unit->m_b1;
    
    if (lag != unit->m_lag) {
        unit->m_b1 = lag == 0.f ? 0.f : exp(log001 / (lag * unit->mRate->mSampleRate));
        unit->m_lag = lag;
    }

    y0 = gLocalSticks[unit->m_id].buttonVals[unit->m_index] ? maxval : minval;

    ZOUT0(0) = y1 = y0 + b1 * (y1 - y0);
    unit->m_y1 = zapgremlins(y1);
}
// toggle button
void JoyButton_next_t( JoyButton * unit, int inNumSamples ) {
    float minval = ZIN0(2);
    float maxval = ZIN0(3);
    float lag = ZIN0(4);
    
    float y1 = unit->m_y1;
    float b1 = unit->m_b1;
    
    if (lag != unit->m_lag) {
        unit->m_b1 = lag == 0.f ? 0.f : exp(log001 / (lag * unit->mRate->mSampleRate));
        unit->m_lag = lag;
    }
    
    int currVal = gLocalSticks[unit->m_id].buttonVals[unit->m_index];

    if( currVal == 1 && unit->m_prevVal == 0 )
        unit->m_state = !unit->m_state;
    
    float y0 = unit->m_state ? maxval : minval;

    ZOUT0(0) = y1 = y0 + b1 * (y1 - y0);
    unit->m_prevVal = currVal;
    unit->m_y1 = zapgremlins(y1);
}
void JoyButton_Ctor( JoyButton * unit )
{	
    initJoyGen( unit );
    
    switch( (int)ZIN0(5) ) { // used case incase i want to add more modes
        case 0:
            SETCALC(JoyButton_next);
            JoyButton_next(unit, 1);
            break;
        case 1:
            unit->m_prevVal = 0;
            unit->m_state = 0;
            SETCALC(JoyButton_next_t);
            JoyButton_next_t(unit, 1);  
            break;  
    }
}

void JoyAxis_next( JoyGen * unit, int inNumSamples ) {
    
    float minval = ZIN0(2);
    float maxval = ZIN0(3);
    float warp = ZIN0(4);
    float lag = ZIN0(5);

    float y1 = unit->m_y1;
    float b1 = unit->m_b1;
    
    if (lag != unit->m_lag) {
        unit->m_b1 = lag == 0.f ? 0.f : exp(log001 / (lag * unit->mRate->mSampleRate));
        unit->m_lag = lag;
    }
    float y0 = gLocalSticks[unit->m_id].axisVals[unit->m_index]; 
    if (warp == 0.0) {
        y0 = (maxval - minval) * y0 + minval;
    } else {
        y0 = pow(maxval/minval, y0) * minval;
    }
    ZOUT0(0) = y1 = y0 + b1 * (y1 - y0);
    unit->m_y1 = zapgremlins(y1);
}

void JoyAxis_Ctor( JoyGen * unit )
{	
    SETCALC(JoyAxis_next);
    initJoyGen( unit );
    JoyAxis_next(unit, 1);
}

void JoyHat_next( JoyHat * unit, int inNumSamples ) {

    float lag = ZIN0(11);

    float y0;
    float y1 = unit->m_y1;
    float b1 = unit->m_b1;
    
    if (lag != unit->m_lag) {
        unit->m_b1 = lag == 0.f ? 0.f : exp(log001 / (lag * unit->mRate->mSampleRate));
        unit->m_lag = lag;
    }

    switch( gLocalSticks[unit->m_id].hatVals[unit->m_index] ) {
        case JOYGEN_HAT_CENTERED: 	y0 = ZIN0(2); break;
        case JOYGEN_HAT_UP:		y0 = ZIN0(3); break;
        case JOYGEN_HAT_RIGHT:		y0 = ZIN0(5); break;
        case JOYGEN_HAT_DOWN:		y0 = ZIN0(7); break;
        case JOYGEN_HAT_LEFT:		y0 = ZIN0(9); break;
        case JOYGEN_HAT_RIGHTUP: 	y0 = ZIN0(4); break;
        case JOYGEN_HAT_RIGHTDOWN:	y0 = ZIN0(6); break;
        case JOYGEN_HAT_LEFTUP: 	y0 = ZIN0(10); break;
        case JOYGEN_HAT_LEFTDOWN:	y0 = ZIN0(8); break;
        default: 			y0 = 0.f;
    }
        
    ZOUT0(0) = y1 = y0 + b1 * (y1 - y0);
    unit->m_y1 = zapgremlins(y1);
}
void JoyHat_next_t( JoyHat * unit, int inNumSamples ) {

    float lag = ZIN0(11);

    float y0;
    float y1 = unit->m_y1;
    float b1 = unit->m_b1;
    
    if (lag != unit->m_lag) {
        unit->m_b1 = lag == 0.f ? 0.f : exp(log001 / (lag * unit->mRate->mSampleRate));
        unit->m_lag = lag;
    }
    
    int hatPos = gLocalSticks[unit->m_id].hatVals[unit->m_index];
    int thisPos;
    
    if( hatPos == JOYGEN_HAT_CENTERED ) {
        thisPos = unit->m_prevPos;
    } else {
        thisPos = hatPos;
        unit->m_prevPos = thisPos;
    }
    
    switch( thisPos ) {
        case JOYGEN_HAT_CENTERED: 	y0 = ZIN0(2); break;
        case JOYGEN_HAT_UP:		y0 = ZIN0(3); break;
        case JOYGEN_HAT_RIGHT:		y0 = ZIN0(5); break;
        case JOYGEN_HAT_DOWN:		y0 = ZIN0(7); break;
        case JOYGEN_HAT_LEFT:		y0 = ZIN0(9); break;
        case JOYGEN_HAT_RIGHTUP: 	y0 = ZIN0(4); break;
        case JOYGEN_HAT_RIGHTDOWN:	y0 = ZIN0(6); break;
        case JOYGEN_HAT_LEFTUP: 	y0 = ZIN0(10); break;
        case JOYGEN_HAT_LEFTDOWN:	y0 = ZIN0(8); break;
        default: 			y0 = 0.f;
    }
        
    ZOUT0(0) = y1 = y0 + b1 * (y1 - y0);
    unit->m_y1 = zapgremlins(y1);
}
void JoyHat_Ctor( JoyHat * unit )
{	
    initJoyGen( unit );
    
    switch( (int)ZIN0(12) ) {
        case 0:
            SETCALC(JoyHat_next);
            JoyHat_next(unit, 1);
            break;
        case 1:
            SETCALC(JoyHat_next_t);
            unit->m_prevPos = JOYGEN_HAT_CENTERED;
            JoyHat_next_t(unit, 1);
            break;
    }
}

#if SC_LINUX
void JoyTrackBall_next( JoyGen * unit, int inNumSamples ) {
    // ... maybe there should be a trackballX and trackballY?
    // sdl doesn't support trackball on darwin, and even if it did
    // i don't have a trackball stick to test with, and also, the doco's sparse, 
    // please feel free to implement this for linux if you like
}
void JoyTrackBall_Ctor( JoyGen * unit )
{	
    SETCALC(JoyTrackBall_next);
    initJoyGen( unit );
    JoyTrackBall_next(unit, 1);
}
#endif

////////////////////////////////////////////////////////////////////////////////////////	

// threadfunc
#if JOYGEN_JOYSTICK_API == JOYGEN_SDL_JOYSTICK_API
void * updateJoysticks(void* arg)
{
    const float A_MUL = 1.0f/65535;
    const int A_ADD = 32768;
    for (;;) {
        SDL_JoystickUpdate();
        for( int i = 0; i < gNumJoySticks; ++i ) {
            SDL_Joystick * sdl = gSDLSticks[i];
            JoyStick & loc = gLocalSticks[i];
            
            int nb = loc.numButtons;
            int na = loc.numAxes;
            int nh = loc.numHats;
            
            for( int b = 0; b < nb; ++b )
                loc.buttonVals[b] = SDL_JoystickGetButton( sdl, b );
                
            for( int a = 0; a < na; ++a ) // make unipolar float
                loc.axisVals[a] = (float)((SDL_JoystickGetAxis( sdl, a ) + A_ADD) * A_MUL);
            
            for( int h = 0; h < nh; ++h )
                loc.hatVals[h] = SDL_JoystickGetHat( sdl, h );
                
            // numballs is always 0 on darwin with sdl
            #if SC_LINUX
            int nt = loc.numBalls;
            for( int t = 0; t < nt; ++t ) {
                // ... if you want this you'll have to write it
            }
            #endif
        } 
        #if SC_DARWIN
            usleep(UPDATE_TIME_USECS);
        #elif SC_LINUX
            nanosleep ( &gRequestedTime , &gRemainingTime ) ;      
        #endif
    } // for(;;)
    return 0;
}

bool initialize() {

    uint16 failedOpen = 0;
    
    if( SDL_Init(SDL_INIT_JOYSTICK) == -1 ) {
        Print("JoyGen: Could not initialize the SDL:\n\t%s.\n", SDL_GetError());
        goto failure;
    }
    
    gNumJoySticks = SDL_NumJoysticks();
    
    if( gNumJoySticks <= 0 ) {
        Print("JoyGen: No joysticks have been detected on your system.\n");
        goto failure;
    }
    
    gSDLSticks = (SDL_Joystick**)NRTAlloc( gNumJoySticks * sizeof(SDL_Joystick*) );
    gLocalSticks = (JoyStick*)NRTAlloc( gNumJoySticks * sizeof(JoyStick) );
    
    for( int i = 0; i < gNumJoySticks; ++i ) {
        gSDLSticks[i] = 0;
    }
    
    for( int i = 0; i < gNumJoySticks; ++i ) {
        SDL_Joystick * sdl = gSDLSticks[i] = SDL_JoystickOpen(i);
        JoyStick & loc = gLocalSticks[i];
        if( sdl ) {
            Print("JoyGen: Initialized Joystick id# %d\n", i );
            Print("\tName: %s\n", SDL_JoystickName(i));
            Print("\tNumber of Buttons: %d\n", loc.numButtons = SDL_JoystickNumButtons(sdl));
            Print("\tNumber of Axes: %d\n", loc.numAxes = SDL_JoystickNumAxes(sdl));
            Print("\tNumber of Hats: %d\n", loc.numHats = SDL_JoystickNumHats(sdl));
            #if SC_LINUX
                Print("\tNumber of Balls: %d\n", loc.numBalls = SDL_JoystickNumBalls(sdl));
            #endif
        } else {
            Print("JoyGen: Failure opening joystick# %d, name: %s.\n", i, SDL_JoystickName(i) );
            ++failedOpen;
        }
    }
    
    if( failedOpen >= gNumJoySticks ) goto failure;
    return true;
  failure:
    Print("Sorry, no joysticks were initialized.\n");
    return false;
}
// #endif JOYGEN_JOYSTICK_API == JOYGEN_SDL_JOYSTICK_API
#endif 

////////////////////////////////////////////////////////////////////////////////////////	

extern "C" void load( InterfaceTable * inTable ) {
    
    ft = inTable;

    if( initialize() ) {
    
        DefineSimpleUnit(JoyButton);
        DefineUnit("JoyAxis", sizeof(JoyGen), (UnitCtorFunc)&JoyAxis_Ctor, 0, 0);  
        DefineSimpleUnit(JoyHat);
        
        #if SC_LINUX
            DefineUnit("JoyTrackBall", sizeof(JoyGen), (UnitCtorFunc)&JoyTrackBall_Ctor, 0, 0);    
        #endif
       
        pthread_t joyThread;
        pthread_create (&joyThread, NULL, updateJoysticks, (void*)0);
    }

}

////////////////////////////////////////////////////////////////////////////////////////	



